"""
This module contains the operational data classes.
"""

# %%
# Import required packages
import numpy as np
import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src.gabse.engine import Engine
    from src.gabse.agent import Agent


# %%
class Sensor:
    """
    A class representing a sensor that logs data from an agent over time. The sensor logs the sensory data based on the
    getter list that is fed as arguments when the sensor is added to the schedule.

    Parameters
    ----------
    engine: Engine
        Reference to the simulation engine.
    parent : Agent
        The agent to which the sensor is attached.
    frequency : float
        The frequency at which the sensor logs data.

    Attributes
    ----------
    engine: Engine
        Reference to the simulation engine.
    parent : Agent
        The agent to which the sensor is attached.
    logger : list
        A list to store logged data entries.
    frequency : float
        The frequency at which the sensor logs data.
    """

    # Initializes the sensor with engine reference, parent agent, empty logger, and frequency
    def __init__(self, engine: "Engine", parent: "Agent", frequency: float):
        self.engine = engine
        self.parent = parent
        self.logger = list()
        self.frequency = frequency

    # Logs data entries based on specified getters
    def entry(self, *getters: list):
        """
        Logs a data entry by calling specified getter methods from the parent agent.

        Parameters
        ----------
        getters : list
            A list of names of all the getter method to call.
        """
        entry = {"tick": self.engine.schedule.get_tick()}

        for arg in getters:
            # print(g)
            method = getattr(self.parent, "get_" + arg)
            if not callable(method):
                # Tests if it is a boolean getter
                method = getattr(self.parent, "is_" + arg)
                if not callable(method):
                    continue

            data = method()
            # check if data is numpy array and convert to list
            if isinstance(data, np.ndarray):
                data = (
                    data.tolist()
                )  # to avoid reference issues with mutable data types
            else:
                try:
                    data = copy.copy(
                        data
                    )  # to avoid reference issues with mutable data types
                except Exception:
                    pass

            entry[arg] = data

        self.logger.append(entry)
        # print(self.engine.getTick())

    # Getters
    def get_frequency(self) -> float:
        """
        Returns the logging frequency of the sensor.

        Returns
        -------
        frequency : float
            The log frequency.
        """
        return self.frequency

    def get_logger(self) -> list:
        """
        Returns the data log associated to the sensor.

        Returns
        -------
        logger : list
            The data log.
        """
        return self.logger


# %%
class DataCollector:
    """
    The data collection manager used for collecting and exporting the operational data for a simulation. The export is
    stored in a dictionary.

    Parameters
    ----------
    engine:Engine
        The simulation engine

    Attributes
    ----------
    engine:Engine
        The simulation engine
    repo:dict
        The data repository.
    """

    def __init__(self, engine):
        self.engine = engine
        self.repo = dict()

    def collect_data(self):
        """
        Collects data from all agents' sensors and stores it in the repository.
        """
        for agt in self.engine.context.get_agents():
            self.repo[f"{agt.__class__.__name__} {agt.id}"] = (
                agt.get_sensor().get_logger()
            )

        # print(self.repo)

    def export_data(self):
        """
        Exports the collected data repository.

        Returns
        -------
        repo:dict
            The data repository.

        """
        return self.repo



"""
This module contains the simulation agent class.
"""

from typing import Any

# %%
# Import required packages
import numpy as np
from src.gabse.data import Sensor
from numpy import floating
from numpy.typing import NDArray
from scipy.spatial import cKDTree as _cKDTree


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src.gabse.engine import Engine


# %%
class Agent:
    """
    A class representing an agent in the simulation. An agent will possess a specific behavior that it executes during
    the simulation. These behaviors are expressed using methods. A few standard methods for *Agent-Based Simulation (ABS)*
    are automatically included.

    The agent class is meant to be used as a parent class, i.e., any agent type that is to be used will be a child
    class of *Agent*. The child class then inherits the *Agent* behavior to ensure that it is directly compatible
    with the simulation engine and bring in standard *ABS* agent methods.

    Parameters
    ----------
    engine : Engine
        Reference to the simulation engine.
    position : NDArray[np.float64]
        The 3D position of the agent in the simulation space.

    Attributes
    ----------
    id: int
        Unique identifier for the agent, automatically generated.
    position: np.ndarray
        The 3D position of the agent in the simulation space.
    engine: Engine
        Reference to the simulation engine.
    sensor: Sensor
        The sensor associated with the agent.
    """

    # Static variable to keep track of agent IDs
    _id_counter = 0

    # Initialize agent with unique ID, position, engine reference, and empty sensor
    def __init__(self, engine: "Engine", position: NDArray[np.float64] = None):
        Agent._id_counter += 1
        self.id = Agent._id_counter
        if position is None:
            position = np.array([0, 0, 0])
        self.engine = engine
        self.position = position
        self.sensor = None

    def find_neighbours(self, agents: list, noOfNeighbours: int) -> list | Any:
        """
        Calculates the distance between *self* and a list of *agents*, neighbours, based on Euclidean distance. It then
        filters out based on the number of neighbours to include, minimum one.

        Parameters
        ----------
        agents : list
            A list of agents for which to calculate distance with.
        noOfNeighbours : int
            The number of closest neighbours to include.

        Returns
        -------
        neighbours : list or Any
            A list of nearest agents, or single agent if *noOfNeighbours == 1*
        """
        if not agents:
            return [] if noOfNeighbours != 1 else None

        n = len(agents)
        k = min(noOfNeighbours, n)

        # Try KDTree for large n or repeated queries
        try:
            pos = np.vstack([a.get_position() for a in agents])
            tree = _cKDTree(pos)
            dists, idxs = tree.query(self.get_position(), k=k)
            if k == 1:
                return agents[int(idxs)]
            if np.isscalar(idxs):
                idxs = [int(idxs)]
            else:
                idxs = [int(i) for i in np.atleast_1d(idxs)]
            return [agents[i] for i in idxs]
        except Exception:
            # Get self position
            self_pos = self.get_position()

            # stack positions (shape: (n, dim)) and compute squared distances
            pos = self.engine.context.get_positions_array()

            if pos.size == 0:
                return [] if noOfNeighbours != 1 else None

            # compute squared Euclidean distances
            d2 = np.sum((pos - self_pos) ** 2, axis=1)

            # Return based if only one neighbours requested
            if k == 1:
                return agents[int(np.argmin(d2))]
            if k < n:
                idx_k = np.argpartition(d2, k - 1)[:k]
                idx_sorted = idx_k[np.argsort(d2[idx_k])]
            else:
                idx_sorted = np.argsort(d2)

            return [agents[i] for i in idx_sorted[:k]]

    def check_out_of_bounds(self) -> NDArray[np.float64]:
        """
        Checks if the agent is outside the simulation context and if so moves it to the closest point within the context.

        Returns
        -------
        position : NDArray[np.float64]
            The new position, unchanged if original position is within bounds.
        """
        bounds = np.array(self.engine.context.get_dimensions())

        minValues = bounds[0:3]
        maxValues = bounds[3:]

        return np.clip(self.position, minValues, maxValues)

    def move_position(self, position: NDArray[np.float64]):
        """
        Moves the agent to a new position. It also does a check so that the agent is still within the bounds of the context.

        Parameters
        ----------
        position : NDArray[np.float64]
            The new position where the agent it to be placed.
        """
        self.position = position
        self.position = self.check_out_of_bounds()
        # print(self.position)

        try:
            self.engine.context.mark_dirty()
        except Exception:
            pass

    def move_vector(self, vector: NDArray[np.float64]):
        """
        Moves the agent to a new position based on a move vector. It also does a check so that the agent is still
        within the bounds of the context.

        Parameters
        ----------
        vector : NDArray[np.float64]
            The movement vector.
        """
        self.position += vector
        self.position = self.check_out_of_bounds()
        # print(self.position)

        try:
            self.engine.context.mark_dirty()
        except Exception:
            pass

    # Calculate Euclidean distance between two agents
    def get_distance(self, agent2: "Agent") -> floating[Any]:
        """
        Calculates the Euclidean distance between two points.

        Parameters
        ----------
        self : Agent
            The first point.
        agent2 : Agent
            The second point.

        Returns
        -------
        dist : floating[Any]
            The distance
        """
        return np.linalg.norm(self.get_position() - agent2.get_position())

    # def find_shortest_path(self, network):
    #    pass

    # Add sensor to agent
    def add_sensor(self, sensor: Sensor):
        """
        Adds a sensor to the agent.

        Parameters
        ----------
        sensor : Sensor
            The sensor to be added.
        """
        self.sensor = sensor

    # Getters and Setters
    def get_sensor(self) -> Sensor:
        """
        Gets a sensor from the agent.

        Returns
        -------
        sensor : Sensor
            The sensor.
        """
        return self.sensor

    def get_position(self) -> NDArray[np.float64]:
        """
        Gets the position of the agent.

        Returns
        -------
        position : NDArray[np.float64]
            The position of the agent.
        """
        return self.position

    def set_position(self, position: NDArray[np.float64]):
        """
        Sets the position of the agent.

        Parameters
        ----------
        position : NDArray[np.float64]
            The position.
        """
        self.position = position



"""
This module contains the simulation scheduling classes.
"""

# %%
# Import required packages
from sortedcontainers import SortedList
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src.gabse.agent import Agent


# %%
class Action:
    """
    A class representing a scheduled action in the simulation. Each behavior that is to be invoked by an agent is
    scheduled and called using *Action*. It is possible to set an action to be reoccurring by using the *interval*
    parameter.

    Parameters
    ----------
    tick: float
        The simulation tick at which the action is scheduled to occur.
    agent: Agent
        The agent that will perform the action.
    method: str
        The name of the method to be called on the agent.
    args: list, optional
        The arguments to be passed to the method. Can be None, empty list, or "" if no arguments are needed.
    priority: int, optional
        The priority of the action (lower values indicate higher priority). Default is 0.
    interval: float, optional
        The interval for recurring actions. If greater than 0, the action will be rescheduled
        after execution. Default is 0.

    Attributes
    ----------
    tick: float
        The simulation tick at which the action is scheduled to occur.
    agent: Agent
        The agent that will perform the action.
    method: str
        The name of the method to be called on the agent.
    args: list, optional
        The arguments to be passed to the method. Can be None, empty list, or "" if no arguments are needed.
    priority: int, optional
        The priority of the action (lower values indicate higher priority). Default is 0.
    interval: float, optional
        The interval for recurring actions, has to be. If greater than 0, the action will be rescheduled
        after execution. Default is 0.
    """

    def __init__(
        self,
        tick: float,
        agent: "Agent",
        method: str,
        args: list = None,
        priority: int = 0,
        interval: float = 0,
    ):
        self.tick = float(tick)
        self.agent = agent
        self.method = method
        self.args = args
        self.priority = int(priority)
        self.interval = float(interval)

        # check so that interval is greater than zero, if not set to zero.
        if self.interval < 0.0:
            self.interval = 0.0

    def __str__(self):
        return f"Action entry:\ntick: {self.tick}, agent: {self.agent}, method: {self.method}, arguments: {self.args}, priority: {self.priority}, interval: {self.interval}"


# %%
class Schedule:
    """
    A class for managing and executing scheduled actions in the simulation. The core of the schedule is a list where
    all planned actions are stored and executed one by one. The schedule uses a SortedList to maintain order of the
    actions based on tick and priority.

    The schedule uses an event-based stepping approach meaning that it does not use fixed tick steps but instead jumps
    between the ticks of the scheduled actions. This means that the tick can step in various lengths depending on the
    actions. A dynamic tick step approach enables greater flexibility and faster simulations.

    Parameters
    ----------
    tick: float
        The current simulation tick.

    Attributes
    ----------
    schedule: SortedList
        A sorted list of scheduled actions, ordered by tick and priority.
    tick: float
        The current simulation tick.
    """

    # Creates an empty schedule (list) and tick timer, set to zero
    # List is sorted based on tick value of actions and priority
    def __init__(self, tick: float):
        self.schedule = SortedList(key=lambda a: (a.tick, a.priority))
        self.tick = tick

    # Schedule method for adding an action in schedule
    def schedule_action(self, action: Action):
        """
        Schedules an action and places it according to its tick and priority

        Parameters
        ----------
        action : Action
            The action object to be scheduled.
        """
        self.schedule.add(action)

    # Method for stepping forward in simulation
    def step(self) -> float:
        """
        Steps one entry in the schedule. The step method executes the next action entry and, if reoccurring, re-schedules
        it. It also moves the tick forward one instance, can be the same if multiple actions are scheduled at the same tick.

        Returns
        -------
        tick : float
            The new tick
        """
        # If schedule is empty, return current tick
        if not self.schedule:
            return self.tick

        # Checks if previous actions exist and, if so, removes them
        while self.schedule[0].tick < self.tick:
            self.schedule.pop(0)

        # If schedule is empty after removing past actions, return current tick
        if not self.schedule:
            return self.tick

        # Load the first action in schedule
        action = self.schedule[0]

        # Step to next action tick
        self.tick = action.tick

        # Calls action agent method
        method = getattr(action.agent, action.method)

        # print(f"{action.agent} at {self.tick}")
        # print(action.method)
        # print(args)

        # Check and call
        if callable(method):
            if action.args is None or len(action.args) == 0 or action.args == "":
                method()
            else:
                method(*action.args)
            # print(result)  # Output: 1, 2, 3
        else:
            print("Method not found or not callable.")

        # Checks if the action is recurring and, if so, schedules next instance
        if action.interval > 0.0:
            nextAction = Action(
                tick=action.tick + action.interval,
                agent=action.agent,
                method=action.method,
                args=action.args,
                interval=action.interval,
            )
            self.schedule_action(nextAction)

        # Remove the executed action from the schedule
        self.schedule.pop(0)

        # Return current tick
        return self.tick

    #
    def remove_agent_from_list(self, target):
        """
        Removes all actions related to the target agent. This is useful if an agent has become obsolete, e.g. killed.

        Parameters
        ----------
        target : Agent
            The agent whose actions are to be removed.
        """
        self.schedule = SortedList(
            [action for action in self.schedule if action.agent != target],
            key=lambda action: (action.tick, action.priority),
        )

    def get_schedule(self) -> SortedList:
        """
        Gets the schedule.

        Returns
        -------
        schedule: SortedList
            The schedule.
        """
        return self.schedule

    def print_schedule(self):
        """
        Prints all actions in the schedule.
        """
        for action in self.schedule:
            print(action)

    def get_tick(self) -> float:
        """
        Gets the current tick in the schedule.

        Returns
        -------
        tick : float
            The current tick.
        """
        return self.tick

    def get_size(self) -> int:
        """
        Gets the size of the schedule.

        Returns
        -------
        size: int
            The size (length) of the schedule.
        """
        return len(self.schedule)

    def clear_schedule(self):
        """
        Clears the schedule.
        """
        self.schedule.clear()



"""
Copyright (C), 2025, Carl Toller Melén

This is the GABSE (Generic Agent-Based Simulation for Engineering) framework.

"""

# version number
__name__ = "gabse"
__author__ = "Carl Toller Melén"
__version__ = "0.1.3"

from .engine import Engine
from .agent import Agent
from .schedule import Action, Schedule
from .context import Context
from .data import Sensor, DataCollector

__all__ = [
    "Engine",
    "Action",
    "Schedule",
    "Agent",
    "Sensor",
    "Context",
    "DataCollector",
]
